unit LogoMain;

{
Program to show disk usage as a pie-chart

Revision history:

V1.0    1993 Feb 08  First version for Borland's Turbo Pascal for Windows
V2.0.0  1996 Apr 14  Version for Borland's Delphi 2.0
V2.0.2  1996 Apr 16  Pre-load Open dialog with '*.*' file name
                     Add number of files and folders display
V2.0.4  1996 May 26  Add E-mail address to About box
V2.1.0  1996 Jul 20  Add auto-registration as Folder Shell Extension
                     Collect size, allocated and slack simultaneously
                     Add option for displaying in bytes, kilobytes or megabytes
                     Save current displayed folder across option selection
V2.1.2  1996 Oct 20  Don't count folders as files (they are, but...)
V2.1.4  1996 Dec 08  Change to "standard" main form (was ToolWin)
                     Revise logo to mimic PieChart colours better
V2.1.6  1997 Feb 08  Add right-click on pie-chart file properties
V2.1.8  1997 Feb 16  Update PieChart to use Begin/EndUpdate for list box output
                     Use inherited, rather than copied, About box
V2.2.0  1997 Mar 01  Handle NTFS compressed files by recording the actual disk
                     space occupied rather than the nominal file size
V2.2.2  1997 Mar 15  Recompile for updated Pie Chart component
V2.3.0  1997 May 10  Version for Delphi 3.0
V2.3.2  1997 May 18  Updates following H J Harvey suggestions
                       - bitmap Up button
                       - split Status bar display
                     Allow user choice of fonts, saved in registry
                     Make Timer initially enabled, not in Create
                     Make PieChart derivative of MultiColourControl
                     Save PieChart colours in registry
                     Remove small hint status bar panel
                     Update more items on status bar while scanning
                     Make form resizable
                     Make ListBox hint reflect the file name clicked
                     Correct incorrect sequential display of compressed properties
V2.3.4  1997 May 23  Add updated glyph for Folder Back (from HH) and Open buttons
                     Disable Open button while scanning
                     Make status-bar updates more timely
                     Make right-click menu addition optional
                     Make About box read version info from the .EXE file 
                       - with Brad Stowers VersInfo unit
V2.3.6  1997 Sep 13  Version for Delphi 3.01
                     Handle "large fonts" more elegantly
V2.3.8  1997 Oct 06  Add maximum filename length display (for CD-ROM writing)
V2.4.0  1997 Nov 15  Add option to display free disk space (from Peter Jaspers-Fayer)
                     Prevent display of properties for non-existant files
                     Update FileInfo for Delphi bug in FindClose
                     Add "count of files" display
                     Build for use with VCL30.DPL
V2.4.2  1997 Dec 05  Version for Delphi 3.02
                     Handle disks > 2GB
                     Use ShellBrowseForFolder (from Brad Stowers BrowseDr unit)
V2.4.4  1997 Dec 12  Add simple print capability
                     (Requested by: Mark Watkins, mwatkins@telepath.com)
                     Only display NTFS "Compressed" label on NTFS volumes
                     Don't display free bytes if there aren't any, e.g. CD-ROMs
                     Remove Exit speed button, not used in most standard apps
V2.4.6  1998 Jan 11  Only force "compressed" display when showing free space
                     Provide Title for printed output
V2.4.8  1998 Feb 16  Correct registry entry to allow spaces in folder names
V2.5.0  1998 Jun 19  Add Windows 95 OSR 2 large disk info function
                     Add repeated scan option
                     Make the Display Free Disk function enabled by default
V2.5.2  1998 Aug 23  Correct intermittent failure to find updated large disk
                       function on original Windows 95
V2.5.4  1998 Sep 25  Use low-level functions for correct cluster size on FAT32
V2.5.6  1998 Oct 02  Change registry entry to avoid "special" folders, e.g. Desktop
}

interface

uses
  Windows, Classes, Graphics, Forms, Controls, Menus,
  Dialogs, StdCtrls, Buttons, ExtCtrls, ComCtrls, PieChart,
  OptnDlg, FileInfo, MultiColour, Messages, VersInfo, BrowseDr;

type
  TMainForm = class(TForm)
    MainMenu: TMainMenu;
    FileMenu: TMenuItem;
    FileOpenItem: TMenuItem;
    FileExitItem: TMenuItem;
    Help1: TMenuItem;
    AboutItem: TMenuItem;
    SpeedPanel: TPanel;
    OpenBtn: TSpeedButton;
    StatusBar: TStatusBar;
    ListBox1: TListBox;
    Timer1: TTimer;
    BitBtnExit: TBitBtn;
    View1: TMenuItem;
    Options1: TMenuItem;
    Refresh1: TMenuItem;
    N2: TMenuItem;
    cbxUnits: TComboBox;
    Label1: TLabel;
    cbxDisplay: TComboBox;
    Label2: TLabel;
    lblCompressed: TLabel;
    BitBtnUp: TBitBtn;
    BitBtnCancel: TBitBtn;
    PieChart1: TPieChart;
    Timer2: TTimer;
    VersionInfoResource1: TVersionInfoResource;
    lblMaxFileNameLength: TLabel;
    BrowseDirectoryDlg1: TBrowseDirectoryDlg;
    Print1: TMenuItem;
    PrintBtn: TSpeedButton;
    TimerRepeatScan: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FileExit(Sender: TObject);
    procedure FileOpen(Sender: TObject);
    procedure About(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure PieChart1DblClick(Sender: TObject);
    procedure ListBox1DblClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure BitBtnUpClick(Sender: TObject);
    procedure Options1Click(Sender: TObject);
    procedure Refresh1Click(Sender: TObject);
    procedure cbxUnitsChange(Sender: TObject);
    procedure cbxDisplayChange(Sender: TObject);
    procedure PieChart1RightClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ListBox1Click(Sender: TObject);
    procedure Timer2Timer(Sender: TObject);
    procedure FilePrint(Sender: TObject);
    procedure TimerRepeatScanTimer(Sender: TObject);
  private
    folder_list: TFolderList;
    current_folder: TFolderList;
    scanning: boolean;
    stop_requested: boolean;
    display_mode: Tdisplay_mode;
    display_units: Tdisplay_units;
    display_compressed: Tdisplay_compressed;
    suggested_min_x: integer;
    suggested_min_y: integer;
    folder_being_scanned: string;
    folder_menu_right_click: boolean;
    display_free_disk: boolean;
    scan_is_repeated: boolean;
    scan_repeat_interval: integer;
    root: string;
    procedure scan_tree;
    procedure display_list (const list: TFolderList);
    procedure handle_double_click (entry: TFolderData);
  protected
    procedure GetMinMaxInfo (var info: TWMGetMinMaxInfo);  message WM_GETMINMAXINFO;
  public
    { Public declarations }
  end;

procedure set_status_text (const s: string);  stdcall;

var
  MainForm: TMainForm;


implementation

uses
  SysUtils, LogoStrs, Registry, FilProps, AbtBox, Printers;

const
  profile_path = '\Software\David J Taylor\ShowMan';
  ListboxFontPrefix = 'ListboxFont';
  PiechartFontPrefix = 'PieChartFont';
  KeySize = '.Size';
  KeyName = '.Name';
  KeyBold = '.Bold';
  KeyItalic = '.Italic';
  KeyColor = '.Color';
  KeyDisplayCompressed = 'Display compressed';
  KeyDisplayMode = 'Display mode';
  KeyDisplayUnits = 'Display units';
  KeyPiechartColours = 'PieChartColours';
  KeyFolderRightClick = 'RightClickFolder';
  KeyDisplayFreeDisk = 'DisplayFreeDiskSpace';
  KeyRepeatedScan = 'RepeatedScan';
  KeyRepeatInterval = 'RepeatInterval';

const
  minutes_to_ms = 60 * 1000;

{$R *.DFM}

var
  under_nt: boolean;

procedure set_status_text (const s: string);
// this procedure is called for each folder scanned
begin
  MainForm.folder_being_scanned := s;
end;

procedure read_font_props (const prefix: string;  font: TFont);
begin
  with TRegIniFile.Create (profile_path) do
    begin
    font.Name := ReadString ('', prefix + KeyName, font.Name);
    font.Size := ReadInteger ('', prefix + KeySize, font.Size);
    font.Color := ReadInteger ('', prefix + KeyColor, font.Color);
    if ReadBool ('', prefix + KeyBold, fsBold in font.Style)
      then font.Style := font.Style + [fsBold]
      else font.Style := font.Style - [fsBold];
    if ReadBool ('', prefix + KeyItalic, fsItalic in font.Style)
      then font.Style := font.Style + [fsItalic]
      else font.Style := font.Style - [fsItalic];
    Free;
    end;
end;

procedure enable_folder_shell_command (enable: boolean);
const
  classes: array [0..1] of string = ('\Directory', '\Drive');
var
  class_num: integer;
  base_class: string;
begin
  // Code for Windows NT 4 and 95 - add Folder right-click menu item
  // Add appropriate pointer to the .EXE and a parameter string placeholder
  with TRegistry.Create do           // use returned TRegistry object
    begin
    RootKey := HKEY_CLASSES_ROOT;
    // First, get rid of the old command, if it exists.
    if OpenKey ('\Folder\shell', False) then
      begin
      DeleteKey ('\Folder\shell\Usage Pie Chart\command');
      DeleteKey ('\Folder\shell\Usage Pie Chart');
      end;

    // Now, insert or remove the new format key as required.
    // Note that it occurs for two classes, HKCR\Directory and HKCR\Drive
    for class_num := 0 to 1 do
      begin
      base_class := classes [class_num] + '\shell';
      if OpenKey (base_class, False) then
        begin
        if enable
        then
          begin
          if OpenKey (base_class + '\UsagePieChart', True) then
            WriteString ('', '&Usage Pie Chart');
          if OpenKey (base_class + '\UsagePieChart\command', True) then
            WriteString ('', Application.ExeName + ' "%1"')
          end
        else
          begin
          DeleteKey (base_class + '\UsagePieChart\command');
          DeleteKey (base_class + '\UsagePieChart');
          end;
        end;
      end;
    Free;
    end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  v: TOSversionInfo;
begin
  folder_list := TFolderList.Create (nil, '.');
  current_folder := nil;
  scanning := False;
  stop_requested := False;
  root := GetCurrentDir;

  // Initialisation code:  note if we are running under Windows NT so that
  // we don't call the GetCompressedFileSize when under Windows 95
  FillChar (v, SizeOf (v), 0);
  v.dwOsVersionInfoSize := SizeOf (v);
  GetVersionEx (v);
  under_nt := v.dwPlatformId = VER_PLATFORM_WIN32_NT;

  // set defaults for the display
  if under_nt
  then display_compressed := show_compressed  // if under NT, enable by default
  else display_compressed := disabled;        // otherwise disabled

  with TRegIniFile.Create (profile_path) do
    begin
    display_compressed := Tdisplay_compressed (
               ReadInteger ('', KeyDisplayCompressed, Ord (display_compressed)));
    display_mode := Tdisplay_mode (     // initial choice is allocated space
               ReadInteger ('', KeyDisplayMode, Ord (show_allocated)));
    display_units := Tdisplay_units (   // and to display it in kilobytes
               ReadInteger ('', KeyDisplayUnits, Ord (show_KB)));
    PieChart1.Colours := ReadString ('', KeyPieChartColours, PieChart1.Colours);
    folder_menu_right_click := ReadBool ('', KeyFolderRightClick, True);
    display_free_disk := ReadBool ('', KeyDisplayFreeDisk, True);
    scan_is_repeated := ReadBool ('', KeyRepeatedScan, False);
    scan_repeat_interval := ReadInteger ('', KeyRepeatInterval, 10);
    Free;
    end;
  cbxUnits.ItemIndex := Ord (display_units);
  cbxDisplay.ItemIndex := Ord (display_mode);

  read_font_props (PiechartFontPrefix, PieChart1.Font);
  read_font_props (ListboxFontPrefix, ListBox1.Font);

  enable_folder_shell_command (folder_menu_right_click);
  suggested_min_x := Width;
  suggested_min_y := Height;
  with TimerRepeatScan do
    begin
    Interval := scan_repeat_interval * minutes_to_ms;
    Enabled := scan_is_repeated;
    end;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  folder_list.Free;
end;

procedure TMainForm.GetMinMaxInfo (var info: TWMGetMinMaxInfo);
begin
  with info.MinMaxInfo.ptMinTrackSize do
    begin
    x := suggested_min_x;
    y := suggested_min_y;
    end;
end;

procedure TMainForm.FileOpen(Sender: TObject);
begin
  if BrowseDirectoryDlg1.Execute then
    begin
    root := BrowseDirectoryDlg1.Selection;
    scan_tree;
    end;
end;

procedure TMainForm.FileExit(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.About(Sender: TObject);
begin
  with AboutBox1 do
    begin
    // blank filename for version resource means use Application.EXEname
    // build up the About Box from the version info in the .EXE file
    ProductName.Caption := VersionInfoResource1.ProductName;
    Version.Caption := 'Version: ' + VersionInfoResource1.FileVersion;
    Copyright.Caption := VersionInfoResource1.LegalCopyright;
    Comments.Caption := VersionInfoResource1.Comments;
    ProgramIcon.Picture.Icon := Application.Icon;
    Caption := 'About ShowMan';
    ShowModal;
    end;
end;

procedure TMainForm.scan_tree;

  // Encapsulate the Windows 98 GetDiskFreeSpaceEx function
  // Also, call low-level Win98 functions to get correct FAT32 cluster size
  function GetDiskFreeBytes (const DirName: string;  var FreeBytes: double;
                             var ClusterSize: integer): boolean;

    const VWIN32_DIOC_DOS_IOCTL = 1;

    type DIOC_REGISTERS =
      record
        reg_EBX: LongInt;
        reg_EDX: LongInt;
        reg_ECX: LongInt;
        reg_EAX: LongInt;
        reg_EDI: LongInt;
        reg_ESI: LongInt;
        reg_Flags: LongInt;
      end;

    const CARRY_FLAG = 1;           // Intel x86 processor status flag

    type DOSDPB =
      packed record
        specialFunc: BYTE;
        devType: BYTE;
        devAttr: Word;
        cCyl: Word;
        mediaType: BYTE;
        cbSec: Word;
        secPerClus: BYTE;
        cSecRes: Word;
        cFAT: BYTE;
        cDir: Word;
        cSec: Word;
        bMedia: BYTE;
        secPerFAT: Word;
        secPerTrack: Word;
        cHead: Word;
        cSecHidden: LongInt;
        cTotalSectors: LongInt;
        reserved: Array [0..5] of BYTE;
      end;

  var
    GetDiskFreeSpaceEx: function (lpDirectoryName: PChar;
                                  var lpFreeBytesAvailableToCaller,
                                      lpTotalNumberOfBytes,
                                      lpTotalNumberOfFreeBytes:
                                  TLargeInteger): bool STDCALL;
    lib: THandle;
    my_bytes: comp;
    total_bytes: comp;
    total_free_bytes: comp;
    Win95: boolean;
    iDrive: integer;
    h: THandle;
    cb: DWORD;
    reg: DIOC_REGISTERS;
    dpb: DOSDPB;
  begin
    Result := False;                         // assume we will fail
    lib := GetModuleHandle ('kernel32');     // get handle of Kernel32
    if lib <> 0 then                         // amazing if it's null!
      begin
      // see if this version of Kernel32 has the updated disk space function
      GetDiskFreeSpaceEx := GetProcAddress (lib, 'GetDiskFreeSpaceExA');
      // check to see if the address is null or not
      if @GetDiskFreeSpaceEx <> nil then           // use updated routine
        begin
        Result := GetDiskFreeSpaceEx (PChar (DirName), TLargeInteger (my_bytes),
                  TLargeInteger (total_bytes), TLargeInteger (total_free_bytes));
        if Result then
          begin
          FreeBytes := total_free_bytes;
          // On Windows 95/8, OSR2, we can try and get the real cluster size
          // for FAT32 volumes.  Try to get the Disk Parameter Block.
          Win95 := GetVersion < 0;
          if Win95 then
            begin
            // On Windows 95, use the technique described in
            // the Knowledge Base article Q125712 and in MSDN under
            // "Windows 95 Guide to Programming", "Using Windows 95
            // features", "Using VWIN32 to Carry Out MS-DOS Functions".
            iDrive := ord (UpCase (DirName [1])) + 1 - Ord ('A');
            h := CreateFile ('\\.\VWIN32', 0, 0, nil, 0, FILE_FLAG_DELETE_ON_CLOSE, 0);

            if h <> INVALID_HANDLE_VALUE then
              begin
              dpb.specialFunc := 0;  // return default type; do not hit disk
              reg.reg_EBX := iDrive;           // BL = drive number (1-based)
              reg.reg_EDX := Integer (@dpb);   // DS:EDX -> DPB
              reg.reg_ECX := $0860;            // CX = Get DPB
              reg.reg_EAX := $440D;            // AX = Ioctl
              reg.reg_Flags := CARRY_FLAG;     // assume failure
              // Make sure both DeviceIoControl and Int 21h succeeded
              if DeviceIoControl (h, VWIN32_DIOC_DOS_IOCTL,   // check IOCTL
                                  @reg, SizeOf (reg),
                                  @reg, SizeOf (reg), cb, nil) and
                 ((reg.reg_Flags and CARRY_FLAG) = 0)         // check Int 21h
              then ClusterSize := dpb.cbSec * dpb.secPerClus;
              CloseHandle (h);
              end;

            end;
          end;
        end;
      end;
  end;

var
  sectors_per_cluster: integer;
  bytes_per_sector: integer;
  free_clusters: integer;
  total_clusters: integer;
  cluster_bytes: integer;
  disk: array [0..3] of char;
  compressed: boolean;
  free_bytes: double;
  compression_supported_on_this_disk: boolean;
  flags: DWORD;
  serial_no: DWORD;
begin
  if scanning then Exit;
  PieChart1.Clear;
  Caption := 'ShowMan';
  // ensure the folder to be scanned ends in a trailing backslash
  if root [Length (root)] <> '\' then root := root + '\';
  // get the C:\ part, no, this code is not UNC compliant....
  StrPLcopy (disk, root, 3);

  // Try and get disk parameters - assume cluster size is one otherwise
  // and assume free space is -1 - a flag value...
  if GetDiskFreeSpace (disk, sectors_per_cluster, bytes_per_sector,
                       free_clusters, total_clusters)
  then
    begin
    cluster_bytes := bytes_per_sector * sectors_per_cluster;
    // Try using the WinNT / Win95, OSR 2 function for large disks
    if not GetDiskFreeBytes (disk, free_bytes, cluster_bytes) then
      begin
      free_bytes := cluster_bytes;                 // convert to floating
      free_bytes := free_bytes * free_clusters;
      end
    end
  else
    begin
    cluster_bytes := 1;
    free_bytes := 0;
    end;

  compression_supported_on_this_disk := False;
  if GetVolumeInformation (disk, nil, 0, nil, serial_no, flags, nil, 0) then
    compression_supported_on_this_disk := (flags and FS_FILE_COMPRESSION) <> 0;

  try
    stop_requested := false;
    BitBtnCancel.Enabled := True;                // allow Cancel
    ActiveControl := BitBtnCancel;               // make the Cancel button active
    OpenBtn.Enabled := False;                    // don't allow any more Opens
    PrintBtn.Enabled := False;                   // or printing
    StatusBar.Panels[0].Text := 'Scanning...';
    StatusBar.Panels[1].Text := '';
    StatusBar.Panels[2].Text := '';
    lblMaxFileNameLength.Caption := '';
    lblMaxFileNameLength.Hint := 'Maximum characters in file name';
    Timer2.Enabled := True;                      // start the status bar updates
    ListBox1.Hint := '';                         // remove any hint
    folder_list.SetFolderName (root);            // mark where we are
    scanning := true;                            // note we're now scanning
    // check for being at root and wanting the free space displayed
    if (Length (root) = 3) and display_free_disk and (free_bytes <> 0) then
      begin
      folder_list.AddFile ('(' + Copy (root, 1, 2) + ' free space)',
                                -free_bytes, -free_bytes);
      if display_compressed <> disabled
        then display_compressed := show_compressed
      end;

    compressed := display_compressed = show_compressed;
    with lblCompressed do
      if not compression_supported_on_this_disk
      then
        Visible := False
      else
        begin
        Visible := True;
        if compressed
        then
          begin
          Caption := 'Compressed sizes';
          Font.Color := clBlue;
          end
        else
          begin
          Caption := 'Uncompressed sizes';
          Font.Color := clBlack;
          end;
        end;

    folder_list.scan (stop_requested, cluster_bytes, compressed, set_status_text);
    display_list (folder_list);                  // show the results
    scanning := false;                           // not scanning now
    stop_requested := false;
    Timer2.Enabled := False;                     // stop status bar updates
    OpenBtn.Enabled := True;                     // allow another Open
    PrintBtn.Enabled := True;                    // allow printing
    BitBtnCancel.Enabled := False;
  finally
  end;
end;

procedure TMainForm.Timer1Timer(Sender: TObject);
begin
  // one-shot code at program start time, scan something to
  // show the user what sort of display to expect
  Timer1.Enabled := False;
  if ParamCount > 0 then root := ParamStr (1);
  scan_tree;
end;

procedure TMainForm.display_list (const list: TFolderList);
var
  str_list: TStringList;
  index: integer;
  size: double;
  total_size: double;
  units: Double;
  s: string;
begin
  if list = nil
    then Exit                        // nothing to do
    else current_folder := list;     // make a note of where we are

  // prepare a string list with the sizes (numeric) and names
  // of the files and folders in this part of the tree
  str_list := TStringList.Create;
  total_size := 0;
  // sort out the user's chosen units for the display
  units := 1;
  case display_units of
    show_KB: units := 1024.0;
    show_MB: units := 1024.0 * 1024.0;
  end;
  for index := 0 to list.Count-1 do
    begin
    size := 0;           // only required to stop the compiler warning
    // choose which number to display
    if display_units = show_count
    then size := TFolderData (list.Objects[index]).Files
    else
      case display_mode of
         show_actual: size := TFolderData (list.Objects[index]).Bytes;
      show_allocated: size := TFolderData (list.Objects[index]).Allocated;
          show_slack: size := TFolderData (list.Objects[index]).Slack;
      end;
    // quantise the display units
    case display_units of
      show_bytes, show_count: Str (size:3:0, s);
      show_KB: Str (size / units:3:1, s);
      show_MB: Str (size / units:3:3, s);
    end;

    str_list.AddObject (s + '   ' + LowerCase (list.Strings[index]),
                        list.Objects[index]);
    total_size := total_size + size;
    end;

  lblMaxFileNameLength.Caption := 'Max path: ' + IntToStr (List.MaxPathLength);
  lblMaxFileNameLength.Hint := 'Maximum characters in file name is ' +
                               IntToStr (List.MaxPathLength) + ' at ' +
                               List.MaxPathString;

  // compute and show the pie-chart
  PieChart1.SetDataAndLabels (str_list);
  str_list.Free;
  // create and display the status line
  with list do
    begin
    case display_mode of
       show_actual: total_size := TotalBytes;
    show_allocated: total_size := TotalAllocated;
        show_slack: total_size := TotalSlack;
    end;
    StatusBar.Panels[0].Text := Format ('Bytes: %1.n', [total_size + 0.0]);
    StatusBar.Panels[1].Text := Format ('Files: %1.n', [TotalFiles + 0.0]);
    StatusBar.Panels[2].Text := Format ('Folders: %1.n', [TotalFolders + 0.0]);
    StatusBar.Panels[3].Text := Format ('%s....', [FolderName]);
    Caption := 'ShowMan - ' + FolderName;    // set a title for the form
    Printer.Title := Caption;                // set a title for any print-out
    // only enable the "Up" button if there's somewhere to go....
    BitBtnUp.Enabled := ParentFolderList <> nil;
    end;
end;

procedure TMainForm.PieChart1DblClick(Sender: TObject);
begin
  handle_double_click (PieChart1.ClickedObject as TFolderData);
end;

procedure TMainForm.ListBox1Click(Sender: TObject);
var
  s: string;
begin
  s := ListBox1.Items.Strings [ListBox1.ItemIndex];
  ListBox1.Hint := Trim (Copy (s, Pos (' ', s), 999));
end;

procedure TMainForm.ListBox1DblClick(Sender: TObject);
var
  entry: TFolderData;
begin
  // determine what item was clicked
  entry := TFolderData (ListBox1.Items.Objects [ListBox1.ItemIndex]);
  handle_double_click (entry);
end;

procedure TMainForm.handle_double_click (entry: TFolderData);
var
  lst: TFolderList;
begin
  // On a double-click, if the item is a sub-folder, go down
  // to it.  If the item a file, then try and go up a level.
  if entry <> nil then
    with entry do
      begin
      lst := SubFolderList;
      if lst <> nil
      then
        display_list (lst)              // we got a sub-folder
      else
        begin
        lst := ParentFolderList;
        if lst <> nil then
          display_list (lst)            // we got a parent folder
        end;
      end;
end;

procedure TMainForm.btnStopClick(Sender: TObject);
begin
  stop_requested := True;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);

  procedure write_font_props (const prefix: string;  const font: TFont);
  begin
    with TRegIniFile.Create (profile_path) do
      begin
      WriteString ('', prefix + KeyName, font.Name);
      WriteInteger ('', prefix + KeySize, font.Size);
      WriteInteger ('', prefix + KeyColor, font.Color);
      WriteBool ('', prefix + KeyBold, fsBold in font.Style);
      WriteBool ('', prefix + KeyItalic, fsItalic in font.Style);
      Free;
      end;
  end;

begin
  stop_requested := True;
  write_font_props (PiechartFontPrefix, PieChart1.Font);
  write_font_props (ListboxFontPrefix, ListBox1.Font);
  with TRegIniFile.Create (profile_path) do
    begin
    WriteInteger ('', KeyDisplayCompressed, Ord (display_compressed));
    WriteInteger ('', KeyDisplayMode, Ord (display_mode));
    WriteInteger ('', KeyDisplayUnits, Ord (display_units));
    WriteString ('', KeyPieChartColours, PieChart1.Colours);
    WriteBool ('', KeyFolderRightClick, folder_menu_right_click);
    WriteBool ('', KeyDisplayFreeDisk, display_free_disk);
    WriteBool ('', KeyRepeatedScan, scan_is_repeated);
    WriteInteger ('', KeyRepeatInterval, scan_repeat_interval);
    Free;
    end;
end;

procedure TMainForm.BitBtnUpClick(Sender: TObject);
// Go up a folder level if there is one
var
  lst: TFolderList;
begin
  if ListBox1.Items.Count <> 0 then
    with TFolderData (ListBox1.Items.Objects [0]) do
    begin
    lst := ParentFolderList;
    if lst <> nil then
      display_list (lst)
    end;
end;

procedure TMainForm.Options1Click(Sender: TObject);
begin
  // Use present setting of DISPLAY_MODE and DISPLAY_UNITS
  // for the initial settings in the options dialog
  OptionsDialog.display_mode := display_mode;
  OptionsDialog.display_units := display_units;
  OptionsDialog.display_compressed := display_compressed;
  OptionsDialog.piechart_font := PieChart1.Font;
  OptionsDialog.piechart_colours := PieChart1.Colours;
  OptionsDialog.listbox_font := ListBox1.Font;
  OptionsDialog.folder_right_click_enabled := folder_menu_right_click;
  OptionsDialog.display_free_disk_enabled := display_free_disk;
  OptionsDialog.repeated_scan := scan_is_repeated;
  OptionsDialog.scan_interval := scan_repeat_interval;

  if OptionsDialog.ShowModal = mrOK then
    begin
    with TimerRepeatScan do
      begin
      scan_is_repeated := OptionsDialog.repeated_scan;
      scan_repeat_interval := OptionsDialog.scan_interval;
      Enabled := scan_is_repeated;
      Interval := scan_repeat_interval * minutes_to_ms;
      end;
    if folder_menu_right_click <> OptionsDialog.folder_right_click_enabled then
      begin
      folder_menu_right_click := OptionsDialog.folder_right_click_enabled;
      enable_folder_shell_command (folder_menu_right_click);
      end;
    if display_free_disk <> OptionsDialog.display_free_disk_enabled then
      begin
      display_free_disk := OptionsDialog.display_free_disk_enabled;
      display_compressed := OptionsDialog.display_compressed;
      display_mode := OptionsDialog.display_mode;
      display_units := OptionsDialog.display_units;
      cbxUnits.ItemIndex := Ord (display_units);
      cbxDisplay.ItemIndex := Ord (display_mode);
      scan_tree;
      end;
    if ListBox1.Font <> OptionsDialog.listbox_font then
      ListBox1.Font := OptionsDialog.listbox_font;
    if PieChart1.Font <> OptionsDialog.piechart_font then
      PieChart1.Font := OptionsDialog.piechart_font;
    if PieChart1.Colours <> OptionsDialog.piechart_colours then
      PieChart1.Colours := OptionsDialog.piechart_colours;
    if display_compressed <> OptionsDialog.display_compressed
    then
      begin
      display_compressed := OptionsDialog.display_compressed;
      scan_tree;
      end
    else
      if (display_mode <> OptionsDialog.display_mode) or
         (display_units <> OptionsDialog.display_units) then
        begin
        // settings have changed, save the new settings and re-display
        display_mode := OptionsDialog.display_mode;
        display_units := OptionsDialog.display_units;
        cbxUnits.ItemIndex := Ord (display_units);
        cbxDisplay.ItemIndex := Ord (display_mode);
        display_list (current_folder);
        end;
    end;
end;

procedure TMainForm.Refresh1Click(Sender: TObject);
begin
  scan_tree;
end;

procedure TMainForm.cbxUnitsChange(Sender: TObject);
// If the units change, simply re-display the current folder.
// No re-scan is required
begin
  case cbxUnits.ItemIndex of
    0: display_units := show_bytes;
    1: display_units := show_KB;
    2: display_units := show_MB;
    3: display_units := show_Count;
  end;
  display_list (current_folder);
end;

procedure TMainForm.cbxDisplayChange(Sender: TObject);
// If the items to display change, simply re-display the current folder.
// No re-scan is required
begin
  case cbxDisplay.ItemIndex of
    0: display_mode := show_actual;
    1: display_mode := show_allocated;
    2: display_mode := show_slack;
  end;
  display_list (current_folder);
end;

procedure TMainForm.PieChart1RightClick(Sender: TObject);
// Handle a PieChart right-click.  If the entry is _not_ a folder, then try
// and display a File Properties dialog box.
var
  file_name: string;
begin
  file_name := PieChart1.ClickedString;
  if file_name [1] <> '\' then                 // not a folder
    with frmFileProperties1 do
      begin
      // build the complete file name
      filename := current_folder.FolderName + file_name;
      if FileExists (filename) then ShowModal;
    end;
end;

procedure TMainForm.FormResize(Sender: TObject);
// Handle the resize command.  Keep the ListBox at a constant size at the right
// edge of the form, and allow the PieChart to fill the remaining space
var
  available_width, available_height: integer;
  diameter: integer;
  h: integer;
begin
  // align the list box a fixed distance from the right hand edge (fixed width)
  ListBox1.Left := ClientWidth - ListBox1.Width - 5;
  // and make the bottom edge of the list box a fixed distance from the bottom
  h := SpeedPanel.Height + StatusBar.Height + BitBtnCancel.Height;
  ListBox1.Height := ClientHeight - h - 24;
  // align the buttons with the bottom left edge of the list box
  // the left positions....
  BitBtnCancel.Left := ListBox1.Left;
  BitBtnUp.Left := BitBtnCancel.Left + BitBtnCancel.Width + 8;
  BitBtnExit.Left := BitBtnUp.Left + BitBtnUp.Width + 8;
  // align buttons vertically
  BitBtnCancel.Top := ListBox1.Top + ListBox1.Height + 7;
  BitBtnUp.Top := BitBtnCancel.Top;
  BitBtnExit.Top := BitBtnCancel.Top;
  lblMaxFileNameLength.Top := BitBtnCancel.Top + BitBtnCancel.Height
                                               - lblMaxFileNameLength.Height;

  // see what space is left for the PieChart - the minimum of width and height
  available_width := ListBox1.Left - 12;    // allow sensible borders
  h := SpeedPanel.Height + StatusBar.Height;
  available_height := ClientHeight - h - 12;
  diameter := available_width;              // assume the width is the smaller
  if available_height < diameter
    then diameter := available_height;      // so actually the height was the limit
  PieChart1.Width := diameter;              // size the PieChart accordingly
  PieChart1.Height := diameter;
end;

procedure TMainForm.Timer2Timer(Sender: TObject);
// this procedure is called three times a second while scanning
// if updated more often, the flicker is more objectionable
begin
  with StatusBar, folder_list do
    begin
    Panels.BeginUpdate;       // freeze the display while updating
    Panels[1].Text := Format ('Files: %1.n', [TotalFiles + 0.0]);
    Panels[2].Text := Format ('Folders: %1.n', [TotalFolders + 0.0]);
    Panels[3].Text := folder_being_scanned;
    Panels.EndUpdate;         // allow the updates to show
    end;
end;

procedure TMainForm.FilePrint(Sender: TObject);
begin
  MainForm.Print;
end;

procedure TMainForm.TimerRepeatScanTimer(Sender: TObject);
begin
  if not scanning then scan_tree;
end;

end.

